# -*- coding: utf-8 -*-
'''
Created on 18.06.2010

@author: adm
'''

import logging
import copy
from enum import Enum
from datetime import datetime
from functools import reduce

log = logging.getLogger("Impact")


class ImpactGroup(object):
    """
    Group of impacts
    """
    def __init__(self):
        self.__impacts = []
        self.check = None
        self.screenMessage = []
        self.receiptMessage = []
        self.clientDisplayMessage = []
        self.coupons = []
        self.counters = []
        self.reports = []

    def __getitem__(self, key):
        return getattr(self, key, None)

    totalsum = property(lambda self: self.getTotalAbsoluteValue())

    def getTotalAbsoluteValue(self):
        """
        Aggregate absolute value of all impacts in group
        """
        total = None
        if self.__impacts:
            total = 0
        for impact in self.__impacts:
            try:
                total += float(impact.absoluteValue)
            except ValueError:
                pass

        return total

    def __repr__(self):
        res = "<ImpactGroup(\n"
        for impact in self.__impacts:
            res += "   %s\n" % impact
        res += ")>"
        return res

    def getImpactsByCode(self, code):
        return list(filter(lambda structImpact: isinstance(structImpact, Impact) and structImpact.impactObject.code == code, self.__impacts))

    def getImpactsAddByCode(self, code):
        return list(filter(lambda structImpact:structImpact.type == ImpactType.ADD.value and structImpact.impactObject.code == code, self.__impacts))

    def getImpactsAddByPositionNumber(self, positionNumber):
        return list(filter(lambda structImpact:structImpact.type == ImpactType.ADD.value and structImpact.impactObject.positionNumber == positionNumber, self.__impacts))

    def getImpactsPureByCodeAndPositionNumber(self, code, positionNumber):
        return list(filter(lambda structImpact: structImpact.type == ImpactType.PURE.value and structImpact.impactObject.code == code and structImpact.impactObject.positionNumber == positionNumber, self.__impacts))

    def getImpactRemoveByPositionNumber(self, positionNumber):
        return list(filter(lambda structImpact: structImpact.type == ImpactType.REMOVE.value and structImpact.impactObject.positionNumber == positionNumber, self.__impacts))

    def getImpactsBranchByPositionNumber(self, positionNumber):
        return list(filter(lambda x: x.type == ImpactType.BRANCH.value and x.impactObject.positionNumber == positionNumber, self.__impacts))

    def addImpact(self, impact):
        """
        Add impact
        """
        log.debug("add impact %s", impact)

        if impact.type == ImpactType.BRANCH.value:
            res = self.getImpactsPureByCodeAndPositionNumber(impact.impactObject.code, impact.impactObject.positionNumber)
            for i in res:
                self.__impacts.remove(i)
                self.distributeImpactToBranches(i, impact)
            self.__impacts.append(impact)

        if impact.type == ImpactType.PURE.value:
            res = self.getImpactsBranchByPositionNumber(impact.impactObject.positionNumber)
            if res:
                for i in res:
                    self.distributeImpactToBranches(impact, i)
            else:
                self.__impacts.append(impact)

        if impact.type == ImpactType.UNITE.value:
            self.__impacts.append(impact)

    def sumImpactQty(self, x, y):
        """
        Sum impact qty
        """
        plus = x
        if isinstance(x, Impact):
            plus = x.qty
        return plus + y.qty

    def distributeImpactToBranches(self, impact, branchImpact):
        remain = impact.absoluteValue
        qty = sum(map(lambda x: x.quantity, branchImpact.branchDetails))
        impacts = []
        for idx, branchDetail in enumerate(branchImpact.branchDetails):
            newimpact = copy.deepcopy(impact)
            newimpact.impactObject.positionNumber = branchDetail.positionNumber
            if idx == len(branchImpact.branchDetails) - 1:
                newimpact.absoluteValue = remain
            else:
                newimpact.absoluteValue = round((impact.absoluteValue / qty) * branchDetail.quantity, 2)
                remain -= newimpact.absoluteValue
            newimpact.qty = branchDetail.quantity
            impacts.append(newimpact)
        log.debug("distributed impacts %s" % impacts)
        self.__impacts.extend(impacts)

    def distributeImpact(self, impact):
        """
        Distribute impact (removed in last step) to receivers
        """
        res = self.getImpactsAddByCode(impact.impactObject.code)

        if len(res) == 0:
            return

        qty = reduce(self.sumImpactQty, res)
        remain = impact.absoluteValue
        distributed = []
        for index in range(0, len(res)):
            newimpact = copy.deepcopy(impact)
            newimpact.impactObject.positionNumber = res[index].impactObject.positionNumber
            newimpact.qty = res[index].qty
            if(index == (len(res) - 1)):
                newimpact.absoluteValue = remain
            else:
                newimpact.absoluteValue = innerRound((impact.absoluteValue / qty) * res[index].qty)
                remain -= newimpact.absoluteValue
            self.__impacts.append(newimpact)
            distributed.append(newimpact)
        log.debug("distribute %s  to %s", impact, res)
        distributedGroup = ImpactGroup()
        distributedGroup.__impacts = distributed
        log.debug("distributed %s", distributedGroup)

    def addImpacts(self, impacts):
        """
        Add list of impact
        """
        list(map(self.addImpact, impacts))

    def getImpacts(self):
        return self.__impacts


class ImpactType(Enum):
    PURE = 'pure'
    ADD = 'add'
    REMOVE = 'remove'
    # addition type
    MODIFY = 'modify'
    BRANCH = 'branch'  # отделение части позиции
    UNITE = 'unite'  # объединение одинаковых позиций


class ImpactMode(Enum):
    AUTOMATIC = 'automatic'
    BYCARD = 'bycard'
    BYDISCKIT = 'bydisckit'
    BYCOUPON = 'bycoupon'


class Impact(object):
    """ Impact """
    cards = []
    coupons = []

    def __init__(self, type=ImpactType.PURE.value, campaign=None, discount=None, order=None, absoluteValue=0,
            impactObject=None, date=None, innerObject=None, qty=None, mode=ImpactMode.AUTOMATIC.value, discountRate=None):
        """
        type - type of impact. ImpactType (ImpactType.PURE,ImpactType.ADD,ImpactType.REMOVE)
        campaign - DicountCampaign
        discount - Discount
        absoluteValue - absolute value of discount
        qty - quantity of goods
        """
        self.campaign = campaign
        self.discount = discount
        self.order = order
        self.absoluteValue = absoluteValue
        self.impactObject = impactObject
        self.date = date if date else datetime.now()
        self.innerObject = innerObject
        self.qty = qty
        self.type = type
        self.branchDetails = None
        self.uniteDetails = None
        self.discountMode = mode
        self.discountRate = discountRate if discountRate else None

    @staticmethod
    def branchImpact(impactObject, branchDetails):
        result = Impact(type=ImpactType.BRANCH.value, impactObject=impactObject)
        result.branchDetails = branchDetails
        return result

    @staticmethod
    def uniteImpact(impactObject, qty, uniteDetails):
        """
        Создает воздействие объединения

        Перемещает количество из позиций указанных в uniteDetails в impactObject
        uniteDetails - словарь вида {<номер позиции откуда переместить>: <кол-во>}
        """
        result = Impact(type=ImpactType.UNITE.value, qty=qty, impactObject=impactObject)
        result.uniteDetails = uniteDetails
        return result

    def __repr__(self):
        if self.type == ImpactType.BRANCH.value:
            return "<%15s(%s,qty=%4s,branchDetails=%s)>" % ("Impact" + ("%s" % self.type).upper(), self.impactObject, self.qty, self.branchDetails)
        elif self.type == ImpactType.UNITE.value:
            return "<%15s(%s,uniteDetails=%s)>" % ("Impact" + ("%s" % self.type).upper(), self.impactObject, self.uniteDetails)
        else:
            return "<%15s(%s,qty=%4s,absoluteValue=%s,mode=%s)>" % ("Impact" + ("%s" % self.type).upper(), self.impactObject, self.qty, self.absoluteValue, ("%s" % self.discountMode).lower())


class ImpactObject(object):
    pass


class ImpactObjectCheck(ImpactObject):
    def __init__(self):
        self.positionNumber = -1
        self.code = None


class ImpactObjectPosition(ImpactObject):
    def __init__(self, positionNumber=None, code=None):
        self.positionNumber = positionNumber
        self.code = code

    def __repr__(self):
        return "<ImpactObjectPosition(%s,%s)>" % (self.positionNumber, repr(self.code))


class InteractionType(Enum):
    ALL = 'all'
    FIRST = 'first'
    LAST = 'last'
    MAX = 'max'
    MIN = 'min'
    MAXFORPOS = 'maxforpos'


def aggregateImpact(impactGroups=None, interactionType=None):
    """
    Aggregate impacts for current hierarchy level

    impactGroups[] - list of ImpactGroup
    interactionType - type of interaction. InteractionType (ALL, FIRST, LAST, MAX, MIN)

    return one ImpactGroup
    """
    if impactGroups is None:
        impactGroups = []
    log = logging.getLogger("AggregateImpact")
    log.debug("aggregate impacts. interactionType=%s", interactionType)
    resultImpact = None
    if interactionType == InteractionType.ALL.value:
        resImpactGroup = ImpactGroup()
        for impactGroup in impactGroups:
            if not impactGroup.getImpacts():
                if resImpactGroup.check is None:
                    resImpactGroup.check = impactGroup.check
                continue
            log.debug("add  to result impactGroup %s", impactGroup)
            resImpactGroup.addImpacts(impactGroup.getImpacts())
            resImpactGroup.check = impactGroup.check
            resImpactGroup.screenMessage.extend(impactGroup.screenMessage)
            resImpactGroup.receiptMessage.extend(impactGroup.receiptMessage)
            resImpactGroup.clientDisplayMessage.extend(impactGroup.clientDisplayMessage)
            resImpactGroup.coupons.extend(impactGroup.coupons)
            resImpactGroup.counters.extend(impactGroup.counters)
            resImpactGroup.reports.extend(impactGroup.reports)

        resultImpact = resImpactGroup
    # InteractionType.LAST сводится к InteractionType.FIRST сменой порядка применения акций(условий) на обратный
    if interactionType in [InteractionType.FIRST.value, InteractionType.LAST.value]:
        if impactGroups:
            filtered = list(filter(lambda impactGroup: impactGroup.getTotalAbsoluteValue() is not None, impactGroups))
            resultImpact = filtered[0] if filtered else ImpactGroup()
    if interactionType == InteractionType.MAX.value:
        if impactGroups:
            filtered = list(filter(lambda impactGroup: impactGroup.getTotalAbsoluteValue() is not None, impactGroups))
            resultImpact = max(filtered, key=lambda impactGroup: impactGroup.getTotalAbsoluteValue()) if filtered else ImpactGroup()
    if interactionType == InteractionType.MIN.value:
        if impactGroups:
            filtered = list(filter(lambda impactGroup: impactGroup.getTotalAbsoluteValue() is not None, impactGroups))
            resultImpact = min(filtered, key=lambda impactGroup: impactGroup.getTotalAbsoluteValue()) if filtered else ImpactGroup()
    if interactionType == InteractionType.MAXFORPOS.value:
        resultImpact = maxForPositions(impactGroups)

    log.debug("Aggregated impacts=%s", resultImpact)
    return resultImpact


def _addNewTmpGroupForPos(tmpImpactGroups, posNum, length):
    """
    Используется в maxForPositions для добавления в tmpImpactGroups нового элемента
    """
    if posNum not in tmpImpactGroups:
        tmpImpactGroups[posNum] = [[0.0, [], x, False] for x in range(length)]


def _addImpactToGroup(tmpImpactGroups, groupNum, posNum, impact):
    """
    Используется в maxForPositions для добавления воздействия в tmpImpactGroups
    """
    try:
        tmpImpactGroups[posNum][groupNum][0] += float(impact.absoluteValue)
    except ValueError:
        pass
    tmpImpactGroups[posNum][groupNum][1].append(impact)


def maxForPositions(impactGroups):
    """
    Выбирает воздействия максимальные для каждой позиции

    Перебирает все группы воздействий, вытаскивает из этих групп воздействия и группирует их по
    позициям и исходным группам. Потом для каждой позиции выбираются воздействия от исходной группы
    с максимальной суммой скидки и добавляются в итоговую группу.

    Воздействие на чек пропорционально распределяется на позиционные воздействия.

    Группы импактов содержащие наборы игнорируются. Этот тип взаимодействия только для
    обычных скидок
    """

    # позиционные воздействия
    # {<номер позиции>: [[<сумма группы на позицию>, [<воздействия группы на позицию>], <номер группы>]]}.
    tmpImpactGroups = {}
    for i in range(0, len(impactGroups)):
        if not impactGroups[i].getImpacts():
            continue

        if any([impact.discountMode == ImpactMode.BYDISCKIT.value for impact in impactGroups[i].getImpacts()]):
            log.warn(u"Взаимодействие 'максимум по позициям' не поддерживает наборы")
            continue

        for impact in impactGroups[i].getImpacts():
            posNum = impact.impactObject.positionNumber
            if posNum == -1:
                # Распределяем скидку на чек по позициям
                discSum = impact.absoluteValue
                disc = discSum / impactGroups[i].check.check.sumb
                newImpact = None
                goodsItems = impactGroups[i].check.getGoodsList()
                for item in goodsItems:
                    roundSum = round(min(item.data["sumb"] * disc, discSum), 2)
                    newImpact = copy.copy(impact)
                    newImpact.impactObject = ImpactObjectPosition(positionNumber=item.data["posnum"], code=item.data["code"])
                    newImpact.qty = item.data["cquant"]
                    newImpact.absoluteValue = roundSum
                    _addNewTmpGroupForPos(tmpImpactGroups, item.data["posnum"], len(impactGroups))
                    _addImpactToGroup(tmpImpactGroups, i, item.data["posnum"], newImpact)
                    discSum -= roundSum
                if discSum > 0.005:
                    roundSum = round(discSum, 2)
                    newImpact.absoluteValue += roundSum
                    tmpImpactGroups[newImpact.impactObject.positionNumber][i][0] += roundSum
            else:
                # Сохраняем сумму и воздействия на позицию
                _addNewTmpGroupForPos(tmpImpactGroups, posNum, len(impactGroups))
                _addImpactToGroup(tmpImpactGroups, i, posNum, impact)

    # Собираем новую группу воздействий
    # Из позиционных групп выбираем максимальные по сумме
    resultImpact = ImpactGroup()
    for impact in impactGroups:
        if impact.check:
            resultImpact.check = impact.check
            break
    usedGroups = set()
    for posNum in tmpImpactGroups:
        maxGroup = max(tmpImpactGroups[posNum], key=lambda group: (group[0]))
        usedGroups.add(maxGroup[2])
        resultImpact.addImpacts(maxGroup[1])
    for i in usedGroups:
        resultImpact.screenMessage.extend(impactGroups[i].screenMessage)
        resultImpact.receiptMessage.extend(impactGroups[i].receiptMessage)
        resultImpact.clientDisplayMessage.extend(impactGroups[i].clientDisplayMessage)
        resultImpact.coupons.extend(impactGroups[i].coupons)
        resultImpact.counters.extend(impactGroups[i].counters)
        resultImpact.reports.extend(impactGroups[i].reports)
    return resultImpact


def innerRound(value):
    return round(value, 2)
